#pragma once

#include <iostream>
#include <vector>
#include <stack>
#include <queue>
#include <map>
#include <algorithm>

#include <fstream>
#include <iomanip>

// デバッグ出力用マクロ
// #define PREPROCESS_DEBUG_OUTPUT
// #define PREPROCESS_PARAMETER_OUTPUT

#ifdef PREPROCESS_DEBUG_OUTPUT
    #define preprocessdbg(str) std::cerr << str << " :line " << __LINE__ << std::endl;
#else
    #define preprocessdbg(str)
#endif

namespace tachibana
{
    class graph
    {
        int n, s, t, limit;
        std::vector<std::pair<int, int>> e;
        bool isdirected;

    public:
        graph(const int numvertices, const std::vector<std::pair<int, int>> &edge, const int start, const int terminal, const int lengthlimit, bool directed) : n(numvertices), e(edge), s(start), t(terminal), limit(lengthlimit), isdirected(directed) {}
        graph(const std::string &filename)
        {
            isdirected = false;
            e.clear();
            std::ifstream in(filename);
            if (!in.is_open())
            {
                std::cerr << "Error: can not open the file " << filename << std::endl;
                return;
            }
            std::string line;
            while (std::getline(in, line))
            {
                // 問題
                if (line[0] == 'p')
                {
                    std::vector<std::string> terms = split(line, ' ');
                    n = std::stoi(terms.at(2));
                }
                // 辺
                else if (line[0] == 'e')
                {
                    std::vector<std::string> terms = split(line, ' ');
                    int u = std::stoi(terms.at(1));
                    int v = std::stoi(terms.at(2));
                    if (u == v) continue;
                    e.emplace_back(u, v);
                }
                else if (line[0] == 'a')
                {
                    std::vector<std::string> terms = split(line, ' ');
                    int u = std::stoi(terms.at(1));
                    int v = std::stoi(terms.at(2));
                    if (u == v) continue;
                    e.emplace_back(u, v);
                    isdirected = true;
                }
                // パス長制限
                else if (line[0] == 'l')
                {
                    std::vector<std::string> terms = split(line, ' ');
                    limit = std::stoi(terms.at(1));
                }
                // 端点
                else if (line[0] == 't')
                {
                    std::vector<std::string> terms = split(line, ' ');
                    s = std::stoi(terms.at(1));
                    t = std::stoi(terms.at(2));
                }
                // コメント
                else if (line[0] == 'c')
                {
                    continue;
                }
            }
        }

        void writeGraph(const std::string &filename)
        {
            int m = e.size();
            char file[1024];
            for (int i = 0; i < (int)filename.size(); ++i)
                file[i] = filename[i];
            file[filename.size()] = '\0';
            FILE *fp = fopen(file, "wt");
            fprintf(fp, "p edge %d %d\n", n, m);
            for (int i = 0; i < m; i++)
            {
                auto [u, v] = e[i];
                fprintf(fp, "%c %d %d\n", (isdirected ? 'a' : 'e'), u, v);
            }
            fprintf(fp, "l %d\n", limit);
            fprintf(fp, "t %d %d\n", s, t);
            fclose(fp);
        }

        void changeIndex(int diffidx)
        {
            s += diffidx;
            t += diffidx;
            for (auto &v : e)
            {
                v.first += diffidx;
                v.second += diffidx;
            }
        }

        int getNumVertices() const { return n; }
        int getNumEdges() const { return e.size(); }
        std::vector<std::pair<int, int>> getEdges() const & { return e; }
        int getStart() const { return s; }
        int getTerminal() const { return t; }
        int getLimit() const { return limit; }
        bool getDirectFlag() const { return isdirected; }

    private:
        std::vector<std::string> split(const std::string &str, char delim)
        {
            std::vector<std::string> res;
            std::string tmp;
            for (char c : str)
            {
                if (c == delim)
                {
                    if (tmp.size() > 0)
                    {
                        res.push_back(tmp);
                        tmp.clear();
                    }
                }
                else
                {
                    tmp.push_back(c);
                }
            }
            if (tmp.size() > 0)
            {
                res.push_back(tmp);
            }
            return res;
        }
    };

    /*
    * block-cut-tree class
    * グラフ上の2点s,tを決めたとき，O(N+M)で全頂点に対して以下の答えを求める．
    * s-tパスであって，頂点vを経由するものが存在するかどうか？
    */
    namespace blockCutTree
    {
        const int NMAX = 2000000;
        int eg[NMAX], dep[NMAX], low[NMAX], gr[NMAX], apid[NMAX], sdis[2 * NMAX], tdis[2 * NMAX];
        std::vector<std::pair<int, int>> g[NMAX];
        std::vector<int> egr[NMAX * 2], vgr[NMAX * 2], bct[NMAX * 2];
        int gnum = 0;

        void dfs2(int now)
        {
            for (auto v : g[now])
            {
                if (dep[v.first] == dep[now] + 1 && eg[v.second] == -1)
                {
                    eg[v.second] = gnum;
                    dfs2(v.first);
                }
            }
        }

        void dfs1(int now, int p)
        {
            int reslow = low[now];
            for (auto v : g[now])
            {
                if (dep[v.first] == -1 && v.first != p)
                {
                    dep[v.first] = dep[now] + 1;
                    low[v.first] = dep[v.first];
                    dfs1(v.first, now);
                }
                if (v.first != p) reslow = std::min(reslow, low[v.first]);
            }
            low[now] = reslow;
            for (auto v : g[now])
            {
                if (dep[v.first] == dep[now] + 1 && v.first != p && dep[now] <= low[v.first])
                {
                    apid[now] = 0;
                    eg[v.second] = gnum;
                    dfs2(v.first);
                    ++gnum;
                }
            }
        };

        // 辺集合，パスの始点sと終点tを渡すと，各頂点についてパスに含まれることがあるかを表すbool配列を返す．
        // 0:含まれない，1:含まれる
        std::vector<bool> deleteVertices(std::vector<std::pair<int, int>> e, int s, int t)
        {
            const int inf = 1 << 30;
            int n = 0;
            for (auto &[u, v] : e) if (u > v) std::swap(u, v);
            std::sort(e.begin(), e.end());
            e.erase(std::unique(e.begin(), e.end()), e.end());
            // std::cerr << "edgenum:" << e.size() << std::endl; 
            for (auto [u, v] : e)
            {
                if (n < u) n = u;
                if (n < v) n = v;
            }
            ++n;
            int siz = e.size();
            for (int i = 0; i < siz; ++i) eg[i] = -1;
            for (int i = 0; i < n; ++i)
            {
                dep[i] = -1;
                low[i] = inf;
                gr[i] = -1;
                apid[i] = -1;
            }
            preprocessdbg("    initialized done")
            for (int i = 0; i < siz; ++i)
            {
                g[e[i].first].emplace_back(e[i].second, i);
                g[e[i].second].emplace_back(e[i].first, i);
            }
            for (int i = 0; i < n; ++i)
            {
                if (dep[i] == -1)
                {
                    dep[i] = 0;
                    low[i] = 0;
                    dfs1(i, -1);
                    bool isisolation = true;
                    for (auto v : g[i])
                    {
                        if (dep[v.first] == 1 && eg[v.second] == -1)
                        {
                            isisolation = false;
                            break;
                        }
                    }
                    if (!isisolation)
                    {
                        dfs2(i);
                        ++gnum;
                    }
                }
            }
            int apcnt = 0;
            for (int i = 0; i < n; ++i)
            {
                if (apid[i] != -1)
                {
                    apid[i] = gnum + apcnt;
                    ++apcnt;
                }
            }
            
            for (int i = 0; i < n; ++i) if (apid[i] != -1) vgr[apid[i]].emplace_back(i);
            for (int i = 0; i < siz; ++i) if (eg[i] != -1) egr[eg[i]].emplace_back(i);
            for (int i = 0; i < gnum; ++i)
            {
                for (int id : egr[i])
                {
                    if (gr[e[id].first] != i)
                    {
                        vgr[i].emplace_back(e[id].first);
                        gr[e[id].first] = i;
                        if (apid[e[id].first] != -1)
                        {
                            int idx = apid[e[id].first];
                            bct[idx].emplace_back(i);
                            bct[i].emplace_back(idx);
                        }
                    }
                    if (gr[e[id].second] != i)
                    {
                        vgr[i].emplace_back(e[id].second);
                        gr[e[id].second] = i;
                        if (apid[e[id].second] != -1)
                        {
                            int idx = apid[e[id].second];
                            bct[i].emplace_back(idx);
                            bct[idx].emplace_back(i);
                        }
                    }
                }
            }
            gnum += apcnt;
            preprocessdbg("    construct block-cut-tree")

            std::vector<bool> ret(n);
            s = (apid[s] == -1 ? gr[s] : apid[s]);
            t = (apid[t] == -1 ? gr[t] : apid[t]);
            for (int i = 0; i < gnum; ++i)
            {
                sdis[i] = inf;
                tdis[i] = inf;
            }
            sdis[s] = 0;
            std::queue<int> q;
            for (q.push(s); !q.empty(); q.pop())
            {
                int now = q.front();
                for (int v : bct[now])
                {
                    if (sdis[v] > sdis[now] + 1)
                    {
                        sdis[v] = sdis[now] + 1;
                        q.push(v);
                    }
                }
            }
            tdis[t] = 0;
            for (q.push(t); !q.empty(); q.pop())
            {
                int now = q.front();
                for (int v : bct[now])
                {
                    if (tdis[v] > tdis[now] + 1)
                    {
                        tdis[v] = tdis[now] + 1;
                        q.push(v);
                    }
                }
            }
            if (sdis[t] < inf) for (int i = 0; i < gnum; ++i) if (sdis[i] + tdis[i] == sdis[t]) for (int v : vgr[i]) ret[v] = true;
            preprocessdbg("    BFS on bct done")
            for (int i = 0; i < n; ++i) std::vector<std::pair<int, int>>().swap(g[i]);
            for (int i = 0; i < gnum; ++i)
            {
                std::vector<int>().swap(egr[i]);
                std::vector<int>().swap(vgr[i]);
                std::vector<int>().swap(bct[i]);
            }
            return ret;
        }
    };


    namespace digraphPreprocess
    {
        struct edge
        {
            int to, revid, cap;
            edge(int t, int r, int c) : to(t), cap(c), revid(r) {}
        };
        const int NMAX = 300000;
        std::vector<int> G[NMAX];
        std::vector<edge> g[NMAX * 2];
        int nowpath[NMAX];
        int sorce, sink, n, start, terminal, nowpathsiz;
        int isused[NMAX], ischecked[NMAX * 2], res[NMAX];
        bool iscolored[2 * NMAX];

        bool findincleasepath(const int s, const int t)
        {
            static int tmp[2 * NMAX];
            int siz = 0;
            auto dfs = [&](auto f, int now) -> bool
            {
                if (now == t) return true;
                if (!ischecked[now]) return false;
                ischecked[now] = 1;
                tmp[siz++] = now;
                for (auto &[to, ri, cp] : g[now]) if (cp > 0 && f(f, to)) return true;
                return false;
            };
            bool res = dfs(dfs, s);
            while (siz > 0) ischecked[tmp[--siz]] = 0;
            return res;
        }

        bool findpath(const int s, const int t)
        {
            static int tmp[NMAX];
            int siz = 0;
            auto dfs = [&](auto f, int now) -> bool
            {
                if (now == t) return true;
                if (isused[now]) return false;
                isused[now] = 2;
                tmp[siz++] = now;
                for (int to : G[now]) if (!isused[to] && f(f, to)) return true;
                return false;
            };
            bool res = dfs(dfs, s);
            while (siz > 0) isused[tmp[--siz]] = 0;
            return res;
        }

        void add_edge(const int s, const int t)
        {
            g[s].emplace_back(t, g[t].size(), 1);
            g[t].emplace_back(s, g[s].size() - 1, 0);
        }

        void dfs2(int now)
        {
            if (iscolored[now]) return;
            iscolored[now] = true;
            for (auto &[to, ri, cp] : g[now]) if (cp > 0 && !iscolored[to]) dfs2(to);
        }

        void dfs1(int now)
        {
            if (now % 2)
            {
                nowpath[nowpathsiz++] = now / 2;
                for (int i = 0; i < g[now].size(); ++i)
                {
                    auto &[to, ri, cp] = g[now][i];
                    if (cp == 0 || res[to / 2] != -1) continue;
                    std::swap(cp, g[to][ri].cap);
                    dfs1(to);
                    std::swap(cp, g[to][ri].cap);
                }
                --nowpathsiz;
            }
            else
            {
                res[now / 2] = 1;
                // flow(now) > 1
                if (!findincleasepath(sorce, now))
                {
                    std::vector<int> tmpid;
                    for (int i = 0; i < 2 * n; ++i) iscolored[i] = 0;
                    iscolored[sorce - 1] = true;
                    iscolored[sink + 1] = true;
                    for (int i = 0; i < nowpathsiz; ++i)
                    {
                        int v = nowpath[i];
                        dfs2(2 * v);
                        if (iscolored[v * 2] && !iscolored[2 * v + 1])
                        {
                            isused[v] = 1;
                            tmpid.emplace_back(v);
                        }
                        if (!iscolored[2 * v + 1]) dfs2(2 * v + 1);
                    }
                    res[now / 2] = findpath(now / 2, terminal);
                    while (!tmpid.empty())
                    {
                        isused[tmpid.back()] = 0;
                        tmpid.pop_back();
                    }
                }
                if (res[now / 2] == 0)
                {
                    for (auto &[to, ri, cp] : g[now])
                    {
                        cp = 0;
                        g[to][ri].cap = 0;
                    }
                    for (auto &[to, ri, cp] : g[now + 1])
                    {
                        cp = 0;
                        g[to][ri].cap = 0;
                    }
                    return;
                }
                auto &[to, ri, cp] = g[now].back();
                std::swap(cp, g[to][ri].cap);
                dfs1(to);
                std::swap(cp, g[to][ri].cap);
            }
        }

        std::vector<bool> flowPreprocess(int vertexnum, std::vector<std::pair<int, int>> e, const int startpoint, const int terminalpoint)
        {
            n = vertexnum;
            start = startpoint;
            terminal = terminalpoint;
            std::sort(e.begin(), e.end());
            e.erase(std::unique(e.begin(), e.end()), e.end());
            for (int i = 0; i < n; ++i)
            {
                isused[i] = 0;
                res[i] = -1;
            }
            for (int i = 0; i < 2 * n; ++i)
            {
                ischecked[i] = 0;
                iscolored[i] = 0;
            }
            for (auto [u, v] : e)
            {
                if (u == v) continue;
                int from = 2 * u + 1, to = 2 * v;
                G[u].emplace_back(v);
                add_edge(from, to);
            }
            for (int i = 0; i < n; ++i)
            {
                if (start == i || terminal == i) continue;
                int from = 2 * i, to = 2 * i + 1;
                add_edge(from, to);
            }
            preprocessdbg("        constructed flow-graph")
            sorce = 2 * start + 1, sink = 2 * terminal;
            res[start] = 1;
            res[terminal] = 1;
            dfs1(sorce);
            std::vector<bool> ans(n, false);
            for (int i = 0; i < n; ++i) if (res[i] == 1) ans[i] = true;
            for (int i = 0; i < n; ++i) std::vector<int>().swap(G[i]);
            for (int i = 0; i < 2 * n; ++i) std::vector<edge>().swap(g[i]);
            return ans;
        }
    };

    class preprocess
    {
        int n, s, t, limit;
        bool isdirected;
        std::vector<std::pair<int, int>> e;

    public:
        preprocess(const graph &g)
        {
            n = g.getNumVertices();
            s = g.getStart();
            t = g.getTerminal();
            limit = g.getLimit();
            e = g.getEdges();
            isdirected = g.getDirectFlag();
        }

        int getNumVertices() {return n;}
        int getNumEdges() {return e.size();}
        int getStart() {return s;}
        int getTerminal() {return t;}
        int getLength() {return limit;}
        std::pair<int, int> getEdge(int id) {return e[id];}

        void deleteVertices()
        {
            init();
            printparam();
            if (n > 0)
            {
                preprocessdbg("distance preprocess start")
                deleteVerticesDistance();
                preprocessdbg("distance preprocess done")
                printparam();
            }
            if (n > 0)
            {
                preprocessdbg("bct preprocess start")
                changeGraph(blockCutTree::deleteVertices(e, s, t));
                preprocessdbg("bct preprocess done")
                printparam();
            }
            if (n > 0)
            {
                preprocessdbg("delete-branch preprocess start")
                deletestbranch();
                preprocessdbg("delete-branch preprocess done")
                printparam();
            }
            if (n > 0 && isdirected)
            {
                preprocessdbg("digraph-flow preprocess start")
                flowPreprocess();
                preprocessdbg("digraph-flow preprocess done")
                printparam();
            }
        }
        graph convertGraph()
        {
            std::vector<std::pair<int, int>> edges(e);
            for (auto &[u, v] : edges)
            {
                ++u;
                ++v;
            }
            return graph(n, edges, s + 1, t + 1, limit, isdirected);
        }
        void printparam()
        {
    #ifdef PREPROCESS_PARAMETER_OUTPUT
            std::cout << "vertices: " << n << ", edges: " << e.size() << ", s:" << s << ", t:" << t << ", lengthlimit:" << limit << ", " << (isdirected ? "directed" : "undirected") << std::endl;;
    #endif
        }

    private:
        void init()
        {
            std::vector<int> vertexValue(e.size() * 2 + 2);
            for (int i = 0; i < (int)e.size(); ++i)
            {
                vertexValue[i * 2] = e[i].first;
                vertexValue[i * 2 + 1] = e[i].second;
            }
            vertexValue[e.size() * 2] = s;
            vertexValue[e.size() * 2 + 1] = t;
            std::sort(vertexValue.begin(), vertexValue.end());
            vertexValue.erase(std::unique(vertexValue.begin(), vertexValue.end()), vertexValue.end());
            n = vertexValue.size();
            for (auto &[u, v] : e)
            {
                int uid = std::lower_bound(vertexValue.begin(), vertexValue.end(), u) - vertexValue.begin();
                int vid = std::lower_bound(vertexValue.begin(), vertexValue.end(), v) - vertexValue.begin();
                u = uid;
                v = vid;
                if (!isdirected && u > v) std::swap(u, v);
            }
            int sid = std::lower_bound(vertexValue.begin(), vertexValue.end(), s) - vertexValue.begin();
            int tid = std::lower_bound(vertexValue.begin(), vertexValue.end(), t) - vertexValue.begin();
            s = sid;
            t = tid;
            limit = std::min(limit, n);
        }

        // 距離制約による削除
        void deleteVerticesDistance()
        {
            const int inf = 1 << 29;
            std::vector<int> diss(n, inf), dist(n, inf);
            std::vector<std::vector<int>> g(n), gr;
            if (isdirected) gr.resize(n);
            for (auto &v : e)
            {
                g[v.first].emplace_back(v.second);
                if (isdirected) gr[v.second].emplace_back(v.first);
                else g[v.second].emplace_back(v.first);
            }
            auto bfs = [](int x, std::vector<std::vector<int>> &G, std::vector<int> &dis) -> void
            {
                dis[x] = 0;
                std::queue<int> q;
                for (q.push(x); !q.empty(); q.pop())
                {
                    int now = q.front();
                    for (auto v : G[now])
                    {
                        if (dis[v] > dis[now] + 1)
                        {
                            dis[v] = dis[now] + 1;
                            q.push(v);
                        }
                    }
                }
            };
            bfs(s, g, diss);
            bfs(t, (isdirected ? gr : g), dist);
            std::vector<bool> used(n);
            for (int i = 0; i < n; ++i) used[i] = (diss[i] + dist[i] <= limit + 1);
            changeGraph(used);
        }

        // 削除できる頂点の情報から，グラフを変換する．
        void changeGraph(const std::vector<bool> &flag)
        {
            std::vector<int> newid(n);
            int numvertices = 0;
            for (int i = 0; i < n; ++i)
            {
                if (!flag[i]) newid[i] = -1;
                else newid[i] = numvertices++;
            }
            int edgenum = 0, deledgecnt = 0;
            for (auto [u, v] : e)
            {
                if (newid[u] != -1 && newid[v] != -1 && u != v) e[edgenum++] = {newid[u], newid[v]};
                else ++deledgecnt;
            }
            while (deledgecnt--) e.pop_back();
            n = numvertices;
            s = newid[s];
            t = newid[t];
            limit = std::min(limit, n);
        }

        void flowPreprocess()
        {
            for (int i = 0; i < 2; ++i)
            {
                auto res = digraphPreprocess::flowPreprocess(n, e, s, t);
                changeGraph(res);
                printparam();
                std::swap(s, t);
                for (auto &[u, v] : e) std::swap(u, v);
            }
        }

        // s,tが次数1の時の頂点削除
        void deletestbranch()
        {
            int newt = t, news = s, cnt = 0;
            std::vector<bool> ret(n, true);
            if (isdirected)
            {
                std::vector<std::vector<int>> g(n), gr(n);
                for (auto [u, v] : e) g[u].emplace_back(v);
                while (g[news].size() == 1)
                {
                    int olds = news;
                    if (g[olds][0] == newt) break;
                    ret[olds] = false;
                    news = g[news][0];
                    g[olds].clear();
                    auto itr = std::find(g[news].begin(), g[news].end(), olds);
                    if (itr != g[news].end()) g[news].erase(itr);
                    ++cnt;
                }
                for (auto [u, v] : e)
                {
                    if (ret[u] == false || ret[v] == false) continue;
                    gr[v].emplace_back(u);
                }
                while (gr[newt].size() == 1)
                {
                    int oldt = newt;
                    if (gr[oldt][0] == news) break;
                    ret[oldt] = false;
                    newt = gr[newt][0];
                    gr[oldt].clear();
                    auto itr = std::find(g[newt].begin(), g[newt].end(), oldt);
                    if (itr != g[newt].end()) g[newt].erase(itr);
                    ++cnt;
                }
            }
            else
            {
                std::vector<std::vector<int>> g(n);
                for (auto [u, v] : e)
                {
                    g[u].emplace_back(v);
                    g[v].emplace_back(u);
                }
                while (g[news].size() == 1)
                {
                    int olds = news;
                    if (g[olds][0] == newt) break;
                    ret[olds] = false;
                    news = g[news][0];
                    g[olds].clear();
                    g[news].erase(std::find(g[news].begin(), g[news].end(), olds));
                    ++cnt;
                }
                while (g[newt].size() == 1)
                {
                    int oldt = newt;
                    if (g[oldt][0] == news) break;
                    ret[oldt] = false;
                    newt = g[newt][0];
                    g[oldt].clear();
                    g[newt].erase(std::find(g[newt].begin(), g[newt].end(), oldt));
                    ++cnt;
                }
            }
            s = news;
            t = newt;
            limit -= cnt;
            changeGraph(ret);
        }
    };
}

// int main()
// {

//     // std::string filename;
//     // std::cout << "filename: ";
//     // std::cin >> filename;
//     // graph g(filename);
//     // preprocess pr(g);
//     // pr.deleteVertices();
//     // graph afg = pr.convertGraph();
//     // std::string outfilename = "testoutput.txt";
//     // afg.writeGraph(outfilename);



//     std::vector<std::string> s;
//     std::string tmp;
//     while (std::cin >> tmp) s.emplace_back(tmp);
//     for (auto t : s)
//     {
//         std::string prefix = "ICGCA2024/instances/directed_graph_public/";
//         tachibana::graph g(prefix + t);
//         int n[2], m[2], l[2];
//         n[0] = g.getNumVertices();
//         m[0] = g.getNumEdges();
//         l[0] = g.getLimit();
//         tachibana::preprocess pr(g);
//         pr.deleteVertices();
//         tachibana::graph afterg = pr.convertGraph();
//         n[1] = afterg.getNumVertices();
//         m[1] = afterg.getNumEdges();
//         l[1] = afterg.getLimit();
//         std::cout << t << std::endl;
//         std::cout << "vertex: " << n[0] << "->" << n[1] << ", ";
//         std::cout << "edge: " << m[0] << "->" << m[1] << ", ";
//         std::cout << "limit: " << l[0] << "->" << l[1] << std::endl;
//         std::cout << "--------------------------------------" << std::endl;

//         prefix = "ICGCA2024/instances/preprocessed/directed/unweighted/";
//         afterg.writeGraph(prefix + t);
//     }


// }