#pragma once

#include <bits/stdc++.h>
#include "../graph.cpp"

// 使い方
// auto node = BS::beamSearch(ビーム幅, 時間制限(ms), グラフG);
// 例：ビーム幅100, 時間制限が10秒のとき、
// auto node = BS::beamSearch(100, 10000, G);
// 頂点順序の取得
// std::vector<int> perm = BS::getVertexPermutation(node);
// 辺順序の取得
// std::vector<pair<int,int>> edgePermutation = BS::getEdgePermutation(node, edge_list);

constexpr int MAX_VERTICES = 4000;
constexpr int BEAM_WIDTH_MIN = 10;

namespace BS {

using namespace std;
using ll = long long;

class TimeKeeper {
private:
    std::chrono::high_resolution_clock::time_point start_time;
    int64_t time_threshold;

public:
    // 時間制限をミリ単位で指定してインスタンスを作る
    TimeKeeper(const int64_t &time_threshold_) 
     : start_time(std::chrono::high_resolution_clock::now()),
       time_threshold(time_threshold_) {}
    
    // インスタンス生成した時から指定した時間制限を超過したか判定する
    bool isTimeOver() const {
        using std::chrono::duration_cast;
        using std::chrono::milliseconds;
        auto diff = std::chrono::high_resolution_clock::now() - this->start_time;
        return duration_cast<milliseconds>(diff).count() >= time_threshold;
    }

    // 経過時間を返す
    int64_t elapsedTime() const {
        using std::chrono::duration_cast;
        using std::chrono::milliseconds;
        auto diff = std::chrono::high_resolution_clock::now() - this->start_time;
        return duration_cast<milliseconds>(diff).count();
    }

    int64_t getLimit() const { return time_threshold; }
};

class State {
public:
    static int S,T,N,M,D;
    static vector<pair<int,int>> edge_list;
    static vector<bitset<MAX_VERTICES>> Gmatrix;
    static vector<int> hash_table;

public:
    State(const Graph& G) {
        S = G.getStart();
        T = G.getTerminal();
        N = G.numVertices();
        M = G.numEdges();
        vector<int> degree(N);
        for (int i = 0; i < M; ++i) {
            edge_list.emplace_back(G.getEdge(i).first);
            ++degree[edge_list.back().first];
            ++degree[edge_list.back().second];
        }
        D = *max_element(degree.begin(),degree.end());
        now = -1;
        lse_score = 0;
        // ave_score = 0;
        used_num = 0;
        used.reset();
        suffix.reset();

        hash_table.resize(N);
        mt19937_64 mt64_random;
        for (int i = 0; i < N; i++) {
            hash_table[i] = mt64_random();
        }
        hash = 0;

        Gmatrix.resize(N);
        for (int i = 0; i < N; ++i) Gmatrix[i].reset();
        for (int i = 0; i < M; ++i) {
            auto&& [u,v] = G.getEdge(i).first;
            Gmatrix[u].set(v);
            Gmatrix[v].set(u);
        }
    }

    State(const State& state) {
        now = state.now;
        used = state.used;
        lse_score = state.lse_score;
        // ave_score = state.ave_score;
        used_num = state.used_num;
        hash = state.hash;
        suffix = state.suffix;
    }
    
    void apply_move(int action) {
        now = action;
        suffix |= Gmatrix[action] & (~used);
        suffix.reset(action);
        int now_bag_size = 2*suffix.count();

        lse_score = logsumexp(lse_score,now_bag_size);
        // ave_score += (N-used_num)*now_bag_size/(((double)N)*(N+1)/2);
        ++used_num;
        used.set(action);
        hash ^= hash_table[action];
    }

    pair<double,ll> try_move(int action) const {
        double next_lse_score = lse_score;
        // double next_ave_score = ave_score;
        ll next_hash = hash ^ hash_table[action];
        auto tmp = suffix;
        tmp |= Gmatrix[action] & (~used);
        tmp.reset(action);

        int now_bag_size = 2*tmp.count();

        next_lse_score = logsumexp(next_lse_score, now_bag_size);
        // next_ave_score += (N-used_num)*now_bag_size/(((double)N)*(N+1)/2);

        return {next_lse_score, next_hash};
    }

    double logsumexp(double a, double b) const {
        double ret = 0.0;
        ret = exp(a-D) + exp(b-D);
        ret = log(ret);
        ret += D;
        return ret;
    }


    bool isDone() const {return used.count() == N;}

    double getScore() const {return min(1e9, lse_score);}

public:
    int now;    // 今回選んだ頂点番号
    bitset<MAX_VERTICES> used; 
    int used_num;
    double lse_score;           // スコア
    // double ave_score;
    // int max_bag_size;   // 最大バッグサイズ
    ll hash;            // ハッシュ
    bitset<MAX_VERTICES> suffix;
};
int State::S;
int State::T;
int State::N;
int State::M;
int State::D;
vector<pair<int,int>> State::edge_list;

vector<int> State::hash_table;
vector<bitset<MAX_VERTICES>> State::Gmatrix;



struct History {
    int action;
    shared_ptr<History> parent;
    History(const int& action_, shared_ptr<History>& parent_) : action(action_), parent(parent_) {}
};
struct Stack {
    shared_ptr<History> head;
    int top() {
        return head->action;
    }

    Stack push(const int action) {
        return Stack({make_shared<History>(action,head)});
    }

    Stack pop() {
        return Stack({head->parent});
    }
};

struct Node {
    State state;
    Stack move_history;

    Node(State& state_) : state(state_) {}

    double get_score() const {return state.getScore();}

    pair<double,ll> calculate(int action) const {return state.try_move(action);}

    void advance(const int action) {state.apply_move(action);}

    bool isDone() const {return state.isDone();}
};

struct TemporaryNode {
    double score;
    ll hash;
    int node_index;
    int action;

    TemporaryNode() {}
    TemporaryNode(double score_, ll hash_, int node_index_, int action_) : score(score_), hash(hash_), node_index(node_index_), action(action_) {}
};


Node beamSearch(int beam_width, const int time_limit, const Graph& G) {
    int max_depth = G.numVertices();
    vector<Node> nodes,next_nodes;
    nodes.reserve(beam_width);
    next_nodes.reserve(beam_width);
    State init_state(G);
    nodes.emplace_back(init_state);
    nodes.back().move_history = Stack{nullptr};
    Node best_node(init_state);
    
    vector<TemporaryNode> temp_nodes;
    temp_nodes.reserve(beam_width * G.numVertices());
    unordered_map<ll,TemporaryNode> fields;

    TimeKeeper timekeeper(time_limit);
    int now_beam_width = beam_width;
    int start_time, end_time;
    start_time = timekeeper.elapsedTime();
    for (int t = 1; t <= max_depth; ++t) {
        temp_nodes.clear();
        fields.clear();

        for (int i = 0,s = nodes.size(); i < s; ++i) {
            for (int action = 0; action < G.numVertices(); ++action) {
                if (nodes[i].state.used[action]) continue;

                auto [score,hash] = nodes[i].calculate(action);

                if (fields.count(hash)) {
                    if (fields[hash].score > score) fields[hash] = {score,hash,i,action};
                }
                else {
                    fields[hash] = {score,hash,i,action};
                }
            }
        }

        for (auto&& [hash,temp] : fields) {
            temp_nodes.emplace_back(temp);
        }

        int node_size = temp_nodes.size();
        if (node_size > now_beam_width) {
            nth_element(temp_nodes.begin(),temp_nodes.begin()+now_beam_width,temp_nodes.end(),[](TemporaryNode& n1, TemporaryNode& n2){
                // if (n1.score == n2.score) return n1.node_index > n2.node_index;
                return n1.score < n2.score;
            });
        }

        for (int i = 0; i < min(node_size, now_beam_width); ++i) {
            int index = temp_nodes[i].node_index;
            next_nodes.emplace_back(nodes[index]);
            next_nodes.back().advance(temp_nodes[i].action);
            next_nodes.back().move_history = nodes[index].move_history.push(temp_nodes[i].action);
        }

        swap(nodes, next_nodes);
        next_nodes.clear();
        next_nodes.shrink_to_fit();

        if (t < max_depth && t%10 == 0) {
            end_time = timekeeper.elapsedTime();
            now_beam_width = (int)round(now_beam_width*sqrt((10.0*(timekeeper.getLimit()-end_time))/(((max_depth-t)/2.0)*(end_time-start_time+1))));
            now_beam_width = max(BEAM_WIDTH_MIN,min(now_beam_width,beam_width));
            start_time = timekeeper.elapsedTime();
        }
    }

    int arg_best = -1;
    double best_score = DBL_MAX;
    for (int i = 0,s = nodes.size(); i < s; ++i) {
        if (nodes[i].get_score() < best_score) {
            arg_best = i;
            best_score = nodes[i].get_score();
        }
    }
    // cerr << arg_best << endl;
    assert(arg_best>=0);
    // cerr << "score:" << best_score << endl;

    return nodes[arg_best];
}


// beam_searchの結果を復元する
vector<pair<int,int>> getEdgePermutation(const Node& result, const vector<pair<int,int>>& edge_list) {
    // cerr << "test" << endl;
    vector<int> ans;
    Stack move_history = result.move_history;
    while (move_history.head) {
        int action = move_history.top();
        ans.emplace_back(action);
        move_history = move_history.pop();
    }
    reverse(ans.begin(),ans.end());
    // cerr << "test" << endl;
    // for (auto i : ans) cerr << i << ' ';
    // cerr << endl;

    vector<vector<int>> Gm(ans.size());
    for (auto&& [u,v] : edge_list) {
        Gm[u].emplace_back(v);
        Gm[v].emplace_back(u);
    }

    vector<int> invAns(ans.size());
    for (int i = 0; i < ans.size(); ++i) invAns[ans[i]] = i;
    vector<int> is_frontier(ans.size());

    vector<bool> used(ans.size());
    vector<pair<int, int>> ret;
    for(int i : ans) {
        vector<int> tmp;
        for (int j : Gm[i]) {
            if (!used[j]) tmp.emplace_back(j);
        }

        sort(tmp.begin(),tmp.end(), [&](int a, int b) -> bool {
            if (is_frontier[a] == is_frontier[b]) return invAns[a] < invAns[b];
            return is_frontier[a] > is_frontier[b];
        });
        for (int j : tmp) {
            if (i < j) ret.emplace_back(i,j);
            else ret.emplace_back(j,i);
            is_frontier[j] = true;
        }

        used[i] = true;
    }

    return ret;
}

vector<int> getVertexPermutation(const Node& result) {
    vector<int> ans;
    Stack move_history = result.move_history;
    while (move_history.head) {
        int action = move_history.top();
        ans.emplace_back(action);
        move_history = move_history.pop();
    }
    reverse(ans.begin(),ans.end());

    return ans;
}

}