#pragma once

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <cassert>

#include <iostream>
#include <fstream>
#include <vector>
#include <queue>
#include <stack>
#include <set>

#include <algorithm>

#include "graph.cpp"

class DiGraph {
private:
    int const INF = 1 << 30;

    int const n;
    int const m;
    int const l;
    int const s;
    int const t;
    std::vector<std::pair<std::pair<int, int>, int>> const edge;
    std::vector<std::vector<std::pair<int, int>>> fgraph;
    std::vector<std::vector<std::pair<int, int>>> rgraph;
    std::vector<std::vector<int>> dist;
    std::vector<int> dist_s;
    std::vector<int> dist_t;
    int onepair;
    int construct;
    int construct_all;

public:
    DiGraph(int n, int m, int l, int s, int t, std::vector<std::pair<std::pair<int, int>, int>> edge) 
        : n(n), m(m), l(l), s(s), t(t), edge(edge) {
        fgraph.resize(n);
        rgraph.resize(n);
        for(auto it = edge.begin(); it != edge.end(); ++it) {
            int u = (*it).first.first;
            int v = (*it).first.second;
            int c = (*it).second;
            fgraph[u].push_back(std::make_pair(v, c));
            rgraph[v].push_back(std::make_pair(u, c));
        }
        onepair = !(s == -1 && t == -1);
        construct = construct_all = false;
    }

    // Floyd-Warshall Algorithm
    void constructDist_all() {
        dist.assign(n, std::vector<int>(n, INF));
        for(int i = 0; i < m; i++) {
            int u = edge[i].first.first;
            int v = edge[i].first.second;
            int c = edge[i].second;
            dist[u][v] = c;
            dist[v][u] = c;
        }
        for(int i = 0; i < n; i++) dist[i][i] = 0;
        for(int k = 0; k < n; k++) {
            for(int i = 0; i < n; i++) {
                for(int j = 0; j < n; j++) {
                    dist[i][j] = std::min(dist[i][j], dist[i][k] + dist[k][j]);
                }
            }
        }
        construct_all = true;
    }

    // Dijkstra's Algorithm
    void constructDist() {
        dist_s.assign(n, INF);
        std::priority_queue<std::pair<int, int>, 
            std::vector<std::pair<int, int>>, 
            std::greater<std::pair<int, int>>> heap_s;
        heap_s.push({dist_s[s] = 0, s});
        while(!heap_s.empty()) {
            auto [d, from] = heap_s.top(); heap_s.pop();
            if(dist_s[from] < d) continue;
            for(const auto [to, cost]: fgraph[from]) {
                if(dist_s[to] > dist_s[from] + cost) {
                    heap_s.push({dist_s[to] = dist_s[from] + cost, to});
                }
            }
        }

        dist_t.assign(n, INF);
        std::priority_queue<std::pair<int, int>, 
            std::vector<std::pair<int, int>>, 
            std::greater<std::pair<int, int>>> heap_t;
        heap_t.push({dist_t[t] = 0, t});
        while(!heap_t.empty()) {
            auto [d, from] = heap_t.top(); heap_t.pop();
            if(dist_t[from] < d) continue;
            for(const auto [to, cost]: rgraph[from]) {
                if(dist_t[to] > dist_t[from] + cost) {
                    heap_t.push({dist_t[to] = dist_t[from] + cost, to});
                }
            }
        }

        construct = true;
    }

    int isOnepair() const {
        return onepair;
    }

    int getDistance(int u, int v) {
        if(construct_all == false) constructDist_all();
        return dist[u][v];
    }

    std::vector<std::pair<std::pair<int, int>, int>> getEdges() const {
        return edge;
    }

    std::pair<std::pair<int, int>, int> getEdge(int e) const {
        return edge[e];
        // int u = edge[e].first.first;
        // int v = edge[e].first.second;
        // int c = edge[e].second;
        // return std::make_pair(std::make_pair(u, v), c);
    }

    int getStart() const {
        return s;
    }

    int getTerminal() const {
        return t;
    }

    int numVertices() const {
        return n;
    }

    int numEdges() const {
        return m;
    }

    int numLength() const {
        return l;
    }

    std::vector<std::pair<int, int>> getForwardNeighbors(int e) const {
        return fgraph[e];
    }

    std::vector<std::pair<int, int>> getReverseNeighbors(int e) const {
        return rgraph[e];
    }

    void printEdges() const {
        for(auto it = edge.begin(); it != edge.end(); ++it) {
            std::cout << (*it).first.first << ", " << (*it).first.second << std::endl;
        }
    }

    void print() const {
        std::cout << "Vertices: " << numVertices() << std::endl;
        std::cout << "Edges: "    << numEdges() << std::endl;
        std::cout << "Length: "   << numLength() << std::endl;
        std::cout << "Terminal: " << getStart() << ", " << getTerminal() << std::endl;
        std::cout << "Edge List: " << std::endl;
        printEdges();
    }

    /*
    void print() const {
        std::cout << "-------- Graph Output --------" << std::endl;
        std::cout << "Vertices: " << numVertices() << std::endl;
        std::cout << "Edges: "    << numEdges() << std::endl;
        std::cout << "Length: "   << numLength() << std::endl;
        std::cout << "Terminal: " << getStart() << ", " << getTerminal() << std::endl;
        std::cout << "Edge List: " << std::endl;
        for(int i = 0; i < n; i++) {
            std::cout << i + 1 << ": ";
            for(int j = 0; j < fgraph[i].size(); j++) {
                std::cout << fgraph[i][j].first << " ";
            }
            std::cout << std::endl;
        }
        std::cout << "------------------------------" << std::endl;
    }
    */

    DiGraph deleteVertices() {
        if(construct == false) constructDist();
        std::vector<int> validVertices(n, true);
        std::vector<int> numVertices(n);
        for(int k = 0; k < n; k++) {
            if(dist_s[k] + dist_t[k] > l) validVertices[k] = false;
        }

        int numValidVertices = 0;
        int numValidEdges = 0;
        int idx = 0;
        for(int i = 0; i < n; i++) {
            if(validVertices[i]) {
                numValidVertices++;
                numVertices[i] = idx++;
            }
        }

        std::vector<std::pair<std::pair<int, int>, int>> res_edge;
        for(int i = 0; i < m; i++) {
            int u = edge[i].first.first;
            int v = edge[i].first.second;
            int c = edge[i].second;
            if(validVertices[u] && validVertices[v]) {
                std::pair<int, int> e = std::make_pair(numVertices[u], numVertices[v]);
                res_edge.push_back(std::make_pair(e, c));
                numValidEdges++;
            }
        }
        assert(numValidEdges == res_edge.size());
        DiGraph res(numValidVertices, numValidEdges, l, numVertices[s], numVertices[t], res_edge);
        return res;
    }

    DiGraph contraction() {
        // multiset
        std::vector<std::multiset<std::pair<int,int>>> fg(n);
        std::vector<std::multiset<std::pair<int,int>>> rg(n);
        for(auto it = edge.begin(); it != edge.end(); ++it) {
            int u = (*it).first.first;
            int v = (*it).first.second;
            int c = (*it).second;
            fg[u].insert(std::make_pair(v, c));
            rg[v].insert(std::make_pair(u, c));
        }

        std::vector<bool> isValidVertices(n, true);
        std::vector<int> numVertices(n);
        for (int i = 0; i < n; i++) {
            if (i == s || i == t) continue;

            std::set<int> adj;
            std::set<int> adj_in;
            std::set<int> adj_out;
            for (auto [u,c] : rg[i]) adj.insert(u), adj_in.insert(u);
            for (auto [v,c] : fg[i]) adj.insert(v), adj_out.insert(v);
            
            // 削除の条件
            if (adj.size() != 2) continue;
            if (fg[i].size() > 2 || rg[i].size() > 2) continue;
            if (fg[i].size() == 2 && rg[i].size() == 2) {
                if (adj_in.size() != 2 || adj_out.size() != 2) continue;   
            }

            // 削除
            // 辺をx->y->zとかならx->zの辺を張る
            for (auto [u,c1] : rg[i]) {
                for (auto [v,c2] : fg[i]) {
                    if (u == v) continue;

                    fg[u].insert(std::make_pair(v,c1+c2));
                    rg[v].insert(std::make_pair(u,c1+c2));
                }
            }

            for (auto [u,c] : rg[i]) {
                fg[u].erase(std::make_pair(i,c));
            }
            for (auto [v,c] : fg[i]) {
                rg[v].erase(std::make_pair(i,c));
            }
            fg[i].clear(); rg[i].clear();

            isValidVertices[i] = false;
        }


        int numValidVertices = 0;
        int numValidEdges = 0;
        int idx = 0;
        for(int i = 0; i < n; i++) {
            if(isValidVertices[i]) {
                numValidVertices++;
                numVertices[i] = idx++;
            }
        }

        std::vector<std::pair<std::pair<int, int>, int>> res_edge;
        for(int u = 0; u < n; u++) {
            for(auto [v, c]: fg[u]) {
                std::pair<int, int> e = std::make_pair(numVertices[u], numVertices[v]);
                res_edge.push_back(std::make_pair(e, c));
                numValidEdges++;
            }
        }
        assert(numValidEdges == res_edge.size());
        DiGraph res(numValidVertices, numValidEdges, l, numVertices[s], numVertices[t], res_edge);
        return res;
    }


    std::vector<bool> contractionVertices() {
        std::vector<std::multiset<std::pair<int,int>>> fg(n);
        std::vector<std::multiset<std::pair<int,int>>> rg(n);
        for(auto it = edge.begin(); it != edge.end(); ++it) {
            int u = (*it).first.first;
            int v = (*it).first.second;
            int c = (*it).second;
            fg[u].insert(std::make_pair(v, c));
            rg[v].insert(std::make_pair(u, c));
        }

        std::vector<bool> isValidVertices(n, true);
        std::vector<int> numVertices(n);

        // 削除
        // 辺をu->i->vとかならu->vの辺を張る
        auto eraseVertex = [&](int i){
            for (auto [u,c1] : rg[i]) {
                for (auto [v,c2] : fg[i]) {
                    if (u == v) continue;

                    fg[u].insert(std::make_pair(v,c1+c2));
                    rg[v].insert(std::make_pair(u,c1+c2));
                }
            }

            for (auto [u,c] : rg[i]) {
                fg[u].erase(std::make_pair(i,c));
            }
            for (auto [v,c] : fg[i]) {
                rg[v].erase(std::make_pair(i,c));
            }
            fg[i].clear(); rg[i].clear();
            isValidVertices[i] = false;
        };
        
        for (int i = 0; i < n; i++) {
            if (i == s || i == t) continue;

            if (fg[i].size() == 1) {
                eraseVertex(i);
                continue;
            }
            else if (rg[i].size() == 1) {
                eraseVertex(i);
                continue;
            }

            std::set<int> adj;
            std::set<int> adj_in;
            std::set<int> adj_out;
            for (auto [u,c] : rg[i]) adj.insert(u), adj_in.insert(u);
            for (auto [v,c] : fg[i]) adj.insert(v), adj_out.insert(v);
            
            // 削除の条件
            if (adj.size() != 2) continue;
            if (fg[i].size() > 2 || rg[i].size() > 2) continue;
            // if (fg[i].size() == 2 && rg[i].size() == 2) {
            //     if (adj_in.size() != 2 || adj_out.size() != 2) continue;   
            // }

            eraseVertex(i);
        }

        return isValidVertices;
    }

    DiGraph contraction2() {
        // multiset
        std::vector<std::multiset<std::pair<int,int>>> fg(n);
        std::vector<std::multiset<std::pair<int,int>>> rg(n);
        for(auto it = edge.begin(); it != edge.end(); ++it) {
            int u = (*it).first.first;
            int v = (*it).first.second;
            int c = (*it).second;
            fg[u].insert(std::make_pair(v, c));
            rg[v].insert(std::make_pair(u, c));
        }

        std::vector<bool> isValidVertices(n, true);
        std::vector<int> numVertices(n);

        // 削除
        // 辺をu->i->vとかならu->vの辺を張る
        auto eraseVertex = [&](int i){
            for (auto [u,c1] : rg[i]) {
                for (auto [v,c2] : fg[i]) {
                    if (u == v) continue;

                    fg[u].insert(std::make_pair(v,c1+c2));
                    rg[v].insert(std::make_pair(u,c1+c2));
                }
            }

            for (auto [u,c] : rg[i]) {
                fg[u].erase(std::make_pair(i,c));
            }
            for (auto [v,c] : fg[i]) {
                rg[v].erase(std::make_pair(i,c));
            }
            fg[i].clear(); rg[i].clear();
            isValidVertices[i] = false;
        };
        
        for (int i = 0; i < n; i++) {
            if (i == s || i == t) continue;

            if (fg[i].size() == 1) {
                eraseVertex(i);
                continue;
            }
            else if (rg[i].size() == 1) {
                eraseVertex(i);
                continue;
            }

            std::set<int> adj;
            std::set<int> adj_in;
            std::set<int> adj_out;
            for (auto [u,c] : rg[i]) adj.insert(u), adj_in.insert(u);
            for (auto [v,c] : fg[i]) adj.insert(v), adj_out.insert(v);
            
            // 削除の条件
            if (adj.size() != 2) continue;
            if (fg[i].size() > 2 || rg[i].size() > 2) continue;
            // if (fg[i].size() == 2 && rg[i].size() == 2) {
            //     if (adj_in.size() != 2 || adj_out.size() != 2) continue;   
            // }

            eraseVertex(i);
        }



        int numValidVertices = 0;
        int numValidEdges = 0;
        int idx = 0;
        for(int i = 0; i < n; i++) {
            if(isValidVertices[i]) {
                numValidVertices++;
                numVertices[i] = idx++;
            }
        }

        std::vector<std::pair<std::pair<int, int>, int>> res_edge;
        for(int u = 0; u < n; u++) {
            for(auto [v, c]: fg[u]) {
                std::pair<int, int> e = std::make_pair(numVertices[u], numVertices[v]);
                res_edge.push_back(std::make_pair(e, c));
                numValidEdges++;
            }
        }
        assert(numValidEdges == res_edge.size());
        DiGraph res(numValidVertices, numValidEdges, l, numVertices[s], numVertices[t], res_edge);
        return res;
    }

    Graph convertUndirectedGraph() {
        std::set<std::pair<std::pair<int, int>, int>> set_edge;
        for(int i = 0; i < m; i++) {
            int u = edge[i].first.first;
            int v = edge[i].first.second;
            int c = edge[i].second;
            if(u > v) std::swap(u, v);
            std::pair<int, int> e = std::make_pair(u, v);
            set_edge.insert(std::make_pair(e, 1));
        }
        std::vector<std::pair<std::pair<int, int>, int>> res_edge;
        for(auto e: set_edge) res_edge.push_back(e);
        sort(res_edge.begin(), res_edge.end());
        Graph res(n, res_edge.size(), l, s, t, res_edge);
        return res;
    }
};
